
// -*- mode: c++; c-basic-offset:4 -*-

// This file is part of libdap, A C++ implementation of the OPeNDAP Data
// Access Protocol.

// Copyright (c) 2002,2003,2005 OPeNDAP, Inc.
// Author: James Gallagher <jgallagher@opendap.org>
//         Reza Nekovei <reza@intcomm.net>
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.

// (c) COPYRIGHT URI/MIT 1994-1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      reza            Reza Nekovei (rnekovei@ieee.org)

// netCDF sub-class implementation for NCByte,...NCGrid.
// The files are patterned after the subcalssing examples
// Test<type>.c,h files.
//
// ReZa 1/12/95

#include "config_nc.h"

static char rcsid[] not_used =
    { "$Id: NCArray.cc 19625 2008-09-17 23:00:57Z jimg $"
    };

#include <cstring>
#include <iostream>
#include <sstream>
#include <algorithm>

// #define DODS_DEBUG 1

#include <Error.h>
#include <InternalErr.h>

#include "Dnetcdf.h"
#include "Dncx.h"
#include "NCArray.h"
#include "NCSequence.h"
#include "nc_util.h"
#include "debug.h"

const string spr = ".";         // structure rename

BaseType *NCArray::ptr_duplicate()
{
    return new NCArray(*this);
}

/** Build an NCArray instance.
    @param n The name of the array.
    @param v Use this variable as a template for type of array elements.
    The variable will be copied, so the caller is responsible for freeing
    storage used by the actual parameter. Also, if the actual parameter
    is an Array, libdap++ code will use the template of that Array as the
    template for this NCArray. */
NCArray::NCArray(const string & n, BaseType * v): Array(n, v)
{}

NCArray::NCArray(const NCArray & rhs): Array(rhs), NCAccess(rhs)
{
    m_duplicate(rhs);
}

NCArray::~NCArray()
{
    DBG(cerr << "delete d_source " << d_source << endl);
    delete d_source;
    d_source = 0;
}

void
NCArray::m_duplicate(const NCArray & nca)
{
    dynamic_cast < NCAccess & >(*this).clone(dynamic_cast <
            const NCAccess & >(nca));
}

NCArray & NCArray::operator=(const NCArray & rhs)
{
    if (this == &rhs)
        return *this;

    dynamic_cast < Array & >(*this) = rhs;

    m_duplicate(rhs);

    return *this;
}


void NCArray::store_projection(const string & proj)
{
    // Scan the projections looking for one about this variable
    string::size_type name_pos = proj.find(name());
    if (name_pos == string::npos)
        return;

    // OK, this variable is in there. At name_pos
    string clause;
    string::size_type end_clause_pos = proj.find(',', name_pos);
    if (end_clause_pos == string::npos)
        clause = proj.substr(name_pos);
    else
        clause = proj.substr(name_pos, end_clause_pos - name_pos);

    // For each bracketed group of numbers, get information about a
    // dimension.
    string::size_type idx = 0;
    string::size_type clause_start = clause.find("[", idx);
    string::size_type clause_end = clause.find("]", idx);
    while (clause_start != string::npos || clause_end != string::npos) {
        string dim_proj =
            clause.substr(clause_start, clause_end - clause_start + 1);

        struct dim_proj_info dpi(dim_proj);
        cl_proj.push_back(dpi);

        idx = clause_end + 1;
        clause_start = clause.find("[", idx);
        clause_end = clause.find("]", idx);
    }

}

string
NCArray::build_constraint(int outtype, const size_t * start,
                          const size_t * edges,
                          const ptrdiff_t * stride) throw(Error)
{
    // Test for bad conversion here, before we may call NCSequence (which
    // cannot run the test).
    if (!is_convertable(outtype))
        throw Error(NC_ECHAR, "Character conversion not supported.");

    // CE always starts with the variable's name

    string expr = name();

    // Get dimension sizes and strings for constraint hyperslab
    ostringstream ce;           // Build CE here.
    int dm = 0;                 // Index to start, edge, stride arrays
    int Zedge = 0;              // Search for zero size edges (no data)

    // If we found an array...
    Array::Dim_iter d;
    vector < dim_proj_info >::iterator cl_iter = cl_proj.begin();
    for (d = dim_begin(); d != dim_end(); ++d, ++dm) {
        // Verify stride argument.
        if (stride != NULL && stride[dm] < 1)
            throw Error(NC_ESTRIDE, "Stride less than 1 (one).");

        // Set defaults:
        // *start         NULL => first corner
        // *edges         NULL => everything following start[]
        // *stride        NULL => unity strides
        long instart = start != NULL ? start[dm] : 0;
        long inedges =
            edges != NULL ? edges[dm] : dimension_size(d) - instart;
        long instep = stride != NULL ? stride[dm] : 1;

        int ext_start;
        int ext_step;
        int ext_stop;
        if (cl_iter != cl_proj.end()) {
            ext_start = (*cl_iter).start;
            ext_step = (*cl_iter).stride;
            ext_stop = (*cl_iter).stop;
            ++cl_iter;
        }
        else {
            ext_start = 0;
            ext_step = 1;
            ext_stop = dimension_size(d) - 1;
        }

        DBG2(cerr << instart << " " << inedges << " " << dm << endl);
        DBG2(cerr << ext_start << " " << ext_step << " " << ext_stop <<
             endl);

        // create constraint expr. by combining the constraint from the
        // command line with the constraint gleaned from the API call.
        int Tstart = ext_start + instart * ext_step;
        int Tstep = instep * ext_step;
        int Tstop =
            (ext_stop <
             ext_start + (instart + (inedges - 1) * instep) * ext_step)
            ? ext_stop : ext_start + (instart +
                                      (inedges - 1) * instep) * ext_step;

        // Check the validity of the array constraints
        if ((instart >= dimension_size(d))
            || (instart < 0) || (inedges < 0))
            throw Error(NC_EINVALCOORDS, "Invalid coordinates.");

        if (instart + inedges > dimension_size(d))
            throw Error(NC_EEDGE,
                        "Hyperslab size exceeds dimension size.");

        // Zero edge found, but first check remaining dimensions for error
        if (inedges == 0)
            Zedge = 1;

        ce << "[" << Tstart << ":" << Tstep << ":" << Tstop << "]";
    }

    if (Zedge == 1)
        throw Error(NC_NOERR, "The constraint would return no data.");

    expr += ce.str();           // Return ce via value-result parameter.
    return expr;
}

bool NCArray::is_convertable(int outtype)
{
    Type intype = var()->type();

    if (((outtype == Ttext) && (intype != dods_str_c)
         && (intype != dods_url_c))
        || ((outtype != Ttext) && (intype == dods_str_c)
            && (outtype != Tvoid))
        || ((outtype != Ttext) && (intype == dods_url_c)
            && (outtype != Tvoid)))
        return false;
    else
        return true;
}

/** Note that this method can ignore the elements parameter since it will
    always be the same size as the nels values (the size of the array). If
    the code gets to this method the original variable in the data source
    (the thing actually sent by the server) is an Array so asking for N
    elements must always return N elements. See NCAccess::extract_values()
    and NCStr::extract_values() for cases where this is not always true.

    @param values Destination buffer
    @param elements Expect this number of elements (i.e., what values can hold)
    @param outtype The netCDF type the client expects to see. */
void NCArray::extract_values(void *values, int /*elements*/, int outtype,
                             const ClientParams & cp) throw(Error)
{
    int nels = length();        // number of elements

    switch (var()->type()) {
    case dods_byte_c:
    case dods_int16_c:
    case dods_uint16_c:
    case dods_int32_c:
    case dods_uint32_c:
    case dods_float32_c:
    case dods_float64_c: {
            char *tmpbufin = 0;
            int bytes = buf2val((void **) & tmpbufin);
            if (bytes == 0)
                throw Error(-1,
                            "Could not read any data from remote server.");

            // Get the netCDF type code for this variable.
            nc_type typep =
                dynamic_cast < NCAccess * >(var())->get_nc_type();
            int rcode =
                convert_nc_type(typep, outtype, nels, tmpbufin, values);
            delete[]tmpbufin;
            tmpbufin = 0;
            if (rcode != NC_NOERR)
                throw Error(rcode,
                            "Error copying values between internal buffers [NCArray::extract_values()]");

            break;
        }                       // End of the case where the array template type is a byte, ..., float64

    case dods_str_c:
    case dods_url_c: {
            // rcode = NC_NOERR;
            char *tbfr = (char *) values;
            string *sa = new string[nels];
            string **spa = &sa;
            buf2val((void **) spa);

            DBG(cerr << "nels: " << nels << endl);
            DBG(for (int i = 0; i < nels; ++i) cerr << sa[i] << endl);

            for (int i = 0; i < nels; i++) {
                unsigned int c;
                for (c = 0;
                     c < sa[i].length() || (sa[i].length() == 0 && c == 0);
                     c++) {
                    *tbfr++ = *(sa[i].c_str() + c);
                }
                if (cp.get_string_as_char_array()) {
                    unsigned int strdim = get_strdim();
                    while (c++ < strdim)
                        *tbfr++ = 0;
                }
            }
            delete[]sa;
            break;
        }                       // End of the case where the array template type is a string or url

    default:
        throw Error(NC_EBADTYPE,
                    string
                    ("The netCDF Client Library cannot access variables of type: ")
                    + dynamic_cast < BaseType * >(this)->type_name()
                    + " [NCArray::extract_values()]");
    }                           // End of the switch
}

nc_type NCArray::get_nc_type() throw(InternalErr)
{
    return dynamic_cast < NCAccess * >(var())->get_nc_type();
}

void NCArray::set_source(BaseType * s) throw(InternalErr)
{
    if (s->type() == dods_array_c)
        throw InternalErr(__FILE__, __LINE__,
                          "Array's source is an Array!");

    d_source = s->ptr_duplicate();
    DBG(cerr << "s " << s << ", d_source " << d_source << endl);
}

/** Arrays are really constructor types. To flateen one, flatten the
    template variable and then append the array's dimensions to each of
    the variable returned as a result.

    @note Four cases: 1. top level array, 2. top level array of structs,
    3. array inside a struct, 4. struct array inside a struct.

    @todo Optimize by not doing all that if the contained variable is
    an atomic type. */
VarList
NCArray::flatten(const ClientParams & cp, const string & parent_name)
{
#if 0
    BaseType *btp = var();
    NCAccess *nca = dynamic_cast < NCAccess * >(btp);
    VarList template_vars = nca->flatten(cp, parent_name);
#else
    VarList template_vars =
        dynamic_cast < NCAccess * >(var())->flatten(cp, parent_name);
#endif
    // At this point we can delete the original template variable (which
    // might be a Structure or Sequence, so there may be several variables
    // in place of the single template). Deleting the template keeps from
    // copying it repeatedly in the while-loop.
    add_var(0);

    VarListIter tv = template_vars.begin();
    VarListIter tv_end = template_vars.end();
    VarList new_vars;
    while (tv != tv_end) {
        // Copy this, cast to Array, load the template, push to list
        NCArray *ap = dynamic_cast < NCArray * >(ptr_duplicate());
        ap->set_translated(true);
        ap->add_var(*tv);
        ap->set_name((*tv)->name());
        new_vars.push_back(ap);
        delete *tv;
        *tv++ = 0;
    }

    return new_vars;
}

BaseType *NCArray::find_child_sequence()
{
    if (var()->type() == dods_sequence_c)
        return var();
    else
        return 0;
}

